#include <netinet/tcp.h>
#include "server.h"

using namespace std;

TLSServer::Sink* TLSServer::sink_ = 0;

static const unsigned SNPRINTF_BUFSZ = 96;

//
// Message implementation
//
Message::Message() {
    reset();
}

void Message::reset() {
    transportType_ = NONE;
    tlsServer_ = 0;
    len_ = 0;
    data_ = 0;
    peer_ = ACE_INET_Addr();
}

Message::Message(const Message& msg) {
    transportType_ = msg.transportType_;
    tlsServer_ = msg.tlsServer_;
    len_ = msg.len_;
    data_ = msg.data_;
}

int Message::sendMessage(unsigned char* buf, unsigned len) {
    int count = 0;
    if(tlsServer_) count++;
    
    if(tlsServer_) {
        ACE_DEBUG((LM_DEBUG, "%D [%P] Sending response to TLS client [%C:%d], size[%d]\n", 
                            tlsServer_->getRemoteIP().c_str(), tlsServer_->getRemotePort(), len)); 
        ACE_HEX_DUMP((LM_DEBUG, (const char*)buf, (size_t)len, "message "));
        return tlsServer_->send(buf, len);
    }
    return -1;
}

//
// TLSServer implementation
//
static const uint32_t BUFFER_SIZE = 1024;
TLSServer::TLSServer() : mdata_(0), thdl_(-1),
              inactivityFlg_(true), localBufferSize_(0), connectionOpened_(false), 
              transportType_(Message::TLS), count_(0), maxRdSize_(BUFFER_SIZE) {
   localBuffer_.reset(new uint8_t[BUFFER_SIZE]);
}

TLSServer::~TLSServer() {
    ACE_TRACE(("TLSServer::~TLSServer()"));
    if(thdl_ != -1) ACE_Reactor::instance()->cancel_timer(thdl_);
}

int TLSServer::initTLS(const char* certificate, const char* privateKey){
  ACE_SSL_Context *context = ACE_SSL_Context::instance();
  if(!context)    return(-1);

  if( (certificate == NULL) && (privateKey == NULL)) return(-1);

  if(certificate) 
           if(context->certificate(certificate, SSL_FILETYPE_PEM) == -1) return(-1);
  if(privateKey)  return(context->private_key(privateKey, SSL_FILETYPE_PEM));

  return(-1);
}

int TLSServer::open(void *ptr) {
    if (super::open(ptr) == -1) return -1;
    
    uint32_t idleTime = 5;
    if(idleTime > 0){
        ACE_Time_Value id(idleTime);
        thdl_ = ACE_Reactor::instance()->schedule_timer(this, 0, id, id);
        ACE_DEBUG((LM_DEBUG, "%D [%P] Inactivity timer started....\n"));
    }

    peer().enable(ACE_NONBLOCK);

    uint32_t s = sizeof(struct sockaddr);

    getpeername(peer().get_handle(), (struct sockaddr *) &stclient_, &s);
    client_.set(&stclient_, sizeof(struct sockaddr_in));

    uint16_t port = ntohs(stclient_.sin_port);
    char ipstr[INET6_ADDRSTRLEN];
    inet_ntop(AF_INET, &stclient_.sin_addr, ipstr, sizeof(ipstr));
    
    ACE_DEBUG((LM_INFO, "%D [%P] New TCP connection from %C:%d on socket %d\n", 
                        ipstr, port, peer().get_handle()));

    int on = 1;
    setsockopt(peer().get_handle(), SOL_SOCKET, SO_KEEPALIVE, 
                &on, sizeof(on));

    uint32_t idle = 10;
    setsockopt(peer().get_handle(), IPPROTO_TCP, TCP_KEEPIDLE, 
                &idle, sizeof(idle));

    uint32_t interval = 10;
    setsockopt(peer().get_handle(), IPPROTO_TCP, TCP_KEEPINTVL, 
                &interval, sizeof(interval));

    uint32_t count = 3;
    setsockopt(peer().get_handle(), IPPROTO_TCP, TCP_KEEPCNT, 
                &count, sizeof(count));
    
    uint32_t size = 65535;
    setsockopt(peer().get_handle(), SOL_SOCKET, 
                 SO_SNDBUF, &size, sizeof(size));
    setsockopt(peer().get_handle(), 
                 SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));

    connectionOpened_ = true;
    return 0;
}

void TLSServer::setSink(TLSServer::Sink* sink) {
    sink_ = sink;
}

int TLSServer::recvMsg() {
    int sz = 0;

    if ((sz =  peer().recv(buff_ + localBufferSize_, 
                           maxRdSize_)) < 1) {
                           //BUFFER_SIZE)) < 1) {

        uint16_t port = ntohs(stclient_.sin_port);
        char ipstr[INET6_ADDRSTRLEN];
        inet_ntop(AF_INET, &stclient_.sin_addr, ipstr, sizeof(ipstr));

        if( (sz == 0) || (ACE_OS::last_error() == EWOULDBLOCK) ) {
            ACE_DEBUG((LM_WARNING, 
                    "%D [%P] Received 0 bytes , TCP client [%C:%d] closed connection\n", 
                    ipstr, port));
            return(0); 
        }

        int lasterrno = ACE_OS::last_error();
        ACE_DEBUG((LM_INFO, "%D [%P] Error reading from %C:%d on socket %d, %C\n", 
                            ipstr, port, peer().get_handle(), 
                            ACE_OS::strerror(lasterrno))); 
        return -1;
    }

    return sz;
}

int TLSServer::handle_timeout( const ACE_Time_Value &ct, const void *val) {
    ACE_UNUSED_ARG (ct);
    // the address of val will never be NULL
    //ACE_UNUSED_ARG (val);
    if(inactivityFlg_ == true) {
       ACE_DEBUG((LM_INFO, "%D [%P] TCP Connection will be closed as connection is inactive for long time.\n"));
       return -1;
    }

    inactivityFlg_ = true;
    return 0;
}

int TLSServer::handle_input(ACE_HANDLE fd) {
    // the address of fd will never be NULL
    //ACE_UNUSED_ARG (fd);
    inactivityFlg_ = false;

#if 0
// Uncomment this for testing the partial message read
// when buffer is not sufficient for reading complete
// message
    count_++;
    if (count_ == 20) {
        buff_ = new uint8_t[200];
        maxRdSize_ = 200;
        count_ = 0;
    }
    else {
        buff_ = new uint8_t[BUFFER_SIZE];
        maxRdSize_ = BUFFER_SIZE;
    }
#endif
    buff_ = new uint8_t[BUFFER_SIZE];
    maxRdSize_ = BUFFER_SIZE;

    if(localBufferSize_ > 0) {
        memcpy(buff_, localBuffer_.get(), localBufferSize_);
    }

    int sz = 0;
    if((sz = recvMsg()) <= 0) {
       return -1;
    }

    if (sz > 0) {
        boost::shared_ptr<Message>mdata(new Message);
        mdata->tlsServer_ = NULL;
        mdata->tlsServer_ = this;
        mdata->transportType_ = transportType_;
        mdata->data_ = buff_;
        mdata->len_  = sz + localBufferSize_;
        localBufferSize_ = 0;
        mdata->peer_ = this->client_;
	
        //Call back function invocation
        int msz = invokeCallback(mdata);
        if(msz == -1) {
            ACE_DEBUG((LM_DEBUG, "%D [%P] Received unexpected TCP Message. Closing connection as per app request\n"));
            return -1;
        }

        if (msz == 0) {
            memcpy((char*)localBuffer_.get(), buff_, sz);
            localBufferSize_ = sz;
        }
    }

    delete buff_;
    return 0;
}

int TLSServer::handle_output(ACE_HANDLE fd) {
    // the address of fd will never be NULL
    //ACE_UNUSED_ARG (fd);
    return 0;
}

int TLSServer::handle_close(ACE_HANDLE fd, ACE_Reactor_Mask mask) {
    ACE_TRACE(("TLSServer::handle_close()"));
    
    uint16_t port = ntohs(stclient_.sin_port);
    char ipstr[INET6_ADDRSTRLEN];
    inet_ntop(AF_INET, &stclient_.sin_addr, ipstr, sizeof(ipstr));
    
    ACE_DEBUG((LM_DEBUG, "%D [%P] TCP connection closed for %C:%d Mask:%d\n", 
                        ipstr, port, mask)); 
    
    connectionOpened_ = false;
    if(mask != ACE_Event_Handler::WRITE_MASK) {
        ACE_DEBUG((LM_DEBUG, "%D [%P] Leaving function:TLSServer::handle_close() mask != WRITE_MASK\n"));
        return super::handle_close(fd, mask);
    }
    return 0;
}

int TLSServer::send(const unsigned char *buf, unsigned len) {
    ACE_TRACE(("TLSServer::send()"));
    int n = 0;

    unsigned count = 0;
    do {
        n = peer().send(buf, len);
        count++;
    }while ((n < 0)&&(count < 5));
    
    if( n < (int)len) {
        return -1;
    }

    return n;
}
